@@ -2,11 +2,11 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
+import json
 import datetime
-from mozpool import config, statemachine, statedriver
-from mozpool.bmm import api as bmm_api
-from mozpool.db import data, logs
-from mozpool.sut import api as sut_api
+from mozpool import config, statemachine, statedriver, async
+from mozpool.bmm import api
+from mozpool.db import exceptions
 import mozpool.lifeguard
 
 
@@ -15,27 +15,25 @@ import mozpool.lifeguard
 
 class DeviceStateMachine(statemachine.StateMachine):
 
-    def __init__(self, device_name):
-        statemachine.StateMachine.__init__(self, 'device', device_name)
+    def __init__(self, device_name, db):
+        statemachine.StateMachine.__init__(self, 'device', device_name, db)
         self.device_name = device_name
 
     def read_state(self):
-        state, timeout, counters = data.get_device_state(self.device_name)
-        return state
+        return self.db.devices.get_machine_state(self.device_name)
 
     def write_state(self, new_state, timeout_duration):
         if timeout_duration is None:
             state_timeout = None
         else:
             state_timeout = datetime.datetime.now() + datetime.timedelta(seconds=timeout_duration)
-        data.set_device_state(self.device_name, new_state, state_timeout)
+        self.db.devices.set_machine_state(self.device_name, new_state, state_timeout)
 
     def read_counters(self):
-        state, timeout, counters = data.get_device_state(self.device_name)
-        return counters
+        return self.db.devices.get_counters(self.device_name)
 
     def write_counters(self, counters):
-        data.set_device_counters(self.device_name, counters)
+        self.db.devices.set_counters(self.device_name, counters)
 
 
 ####
@@ -43,8 +41,7 @@ class DeviceStateMachine(statemachine.StateMachine):
 
 class DeviceLogDBHandler(statedriver.DBHandler):
 
-    object_name = 'device'
-    log_object = logs.device_logs
+    object_type = 'device'
 
 
 class LifeguardDriver(statedriver.StateDriver):
@@ -60,13 +57,20 @@ class LifeguardDriver(statedriver.StateDriver):
     thread_name = 'LifeguardDriver'
     log_db_handler = DeviceLogDBHandler
 
-    def __init__(self, poll_frequency=statedriver.POLL_FREQUENCY):
-        statedriver.StateDriver.__init__(self, poll_frequency)
-        self.imaging_server_id = data.find_imaging_server_id(
-            config.get('server', 'fqdn'))
+    def __init__(self, db, poll_frequency=statedriver.POLL_FREQUENCY):
+        statedriver.StateDriver.__init__(self, db, poll_frequency)
+        self.imaging_server_id = self.db.imaging_servers.get_id(config.get('server', 'fqdn'))
+
+        # set up the BMM API for use by machines
+        self.api = api.API(db)
+
+    def _get_machine(self, machine_name):
+        machine = super(LifeguardDriver, self)._get_machine(machine_name)
+        machine.api = self.api
+        return machine
 
     def _get_timed_out_machine_names(self):
-        return data.get_timed_out_devices(self.imaging_server_id)
+        return self.db.devices.list_timed_out(self.imaging_server_id)
 
 
 ####
@@ -86,19 +90,22 @@ class AcceptPleaseRequests(AcceptBasicPleaseRequests):
     "Mixin to also accept requests that assume a working device"
 
     def on_please_power_cycle(self, args):
+        # clear out any "next image" so we don't accidentally reimage
+        self.db.devices.set_next_image(self.machine.device_name, None, None)
         self.machine.goto_state(pc_power_cycling)
 
     def on_please_image(self, args):
         try:
-            data.get_pxe_config_for_device(self.machine.device_name,
+            # verify it's possible (this raises an exception if not)
+            self.db.devices.get_pxe_config(self.machine.device_name,
                                            args['image'])
-            self.logger.info('writing image %s, boot config %s to db' %
+            self.logger.info('setting next image to %s, boot config to %s' %
                              (args['image'], args['boot_config']))
-            data.set_device_config(self.machine.device_name,
+            self.db.devices.set_next_image(self.machine.device_name,
                                    args['image'],
                                    args['boot_config'])
             self.machine.goto_state(start_pxe_boot)
-        except data.NotFound:
+        except exceptions.NotFound:
             self.logger.error('cannot image device')
 
 
@@ -133,11 +140,11 @@ class PowerCycleMixin(object):
 
     def setup_pxe(self):
         # hook for subclasses
-        bmm_api.clear_pxe(self.machine.device_name)
+        self.machine.api.clear_pxe.run(self.machine.device_name)
 
     def on_entry(self):
-        has_sut_agent = data.device_has_sut_agent(self.machine.device_name)
-        has_relay = data.device_relay_info(self.machine.device_name)
+        has_sut_agent = self.db.devices.has_sut_agent(self.machine.device_name)
+        has_relay = self.db.devices.get_relay_info(self.machine.device_name)
         if has_sut_agent and (not has_relay or
                               (self.machine.increment_counter('sut_attempts') <=
                                self.TRY_RELAY_AFTER_SUT_COUNT)):
@@ -163,7 +170,9 @@ class PowerCycleMixin(object):
             else:
                 mozpool.lifeguard.driver.handle_event(self.machine.device_name,
                                                       'power_cycle_failed', {})
-        sut_api.start_reboot(self.machine.device_name, reboot_initiated)
+        self.setup_pxe()
+        self.logger.info("starting SUT reboot")
+        self.machine.api.sut_reboot.start(reboot_initiated, self.machine.device_name)
 
     def relay_powercycle(self):
         # kick off a power cycle on entry
@@ -172,7 +181,8 @@ class PowerCycleMixin(object):
             if success:
                 mozpool.lifeguard.driver.handle_event(self.machine.device_name, 'power_cycle_ok', {})
         self.setup_pxe()
-        bmm_api.start_powercycle(self.machine.device_name, powercycle_done)
+        self.logger.info("initiating power cycle")
+        self.machine.api.powercycle.start(powercycle_done, self.machine.device_name)
 
     def on_timeout(self):
         self.on_power_cycle_failed({})
@@ -187,9 +197,50 @@ class PowerCycleMixin(object):
     def on_power_cycle_ok(self, args):
         self.machine.clear_counter('power_cycling')
         self.machine.clear_counter('sut_attempts')
+        self.power_cycle_complete()
+
+    def power_cycle_complete(self):
+        """
+        Subclasses can override this to do something more complex than going to
+        power_cycle_complete_state
+        """
         self.machine.goto_state(self.power_cycle_complete_state)
 
 
+class ImagingResultMixin(object):
+
+    def send_imaging_result(self, imaging_result):
+        """
+        Send the given imaging result to Mozpool.  This both inserts the result
+        into the db and POSTs it to the Mozpool server, but does nothing if
+        there is no attached request.
+        """
+        device_name = self.machine.device_name
+        self.db.device_requests.set_result(device_name, imaging_result)
+
+        # find out if this device was requested
+        req_id = self.db.device_requests.get_by_device(device_name)
+        if req_id is None:
+            return
+
+        # next, who do we talk to about this request?
+        try:
+            imaging_svr = self.db.requests.get_imaging_server(req_id)
+        except exceptions.NotFound:
+            return
+
+        # tell mozpool we're finished.  This is a notification, so we really
+        # don't care if it succeeds or not
+        self.logger.info("sending imaging result '%s' to Mozpool" % imaging_result)
+        mozpool_url = 'http://%s/api/request/%d/event/lifeguard_finished/' % (
+                imaging_svr, req_id)
+        def posted(result):
+            if result.status_code != 200:
+                self.logger.warn("got %d from Mozpool" % result.status_code)
+        async.requests.post.start(posted, mozpool_url,
+                data=json.dumps({'imaging_result': imaging_result}))
+
+
 ####
 # Initial and steady states
 
@@ -208,7 +259,6 @@ class unknown(AcceptPleaseRequests, statemachine.State):
     "This device is in an unknown state.  Await instructions."
 
 
-
 @DeviceStateMachine.state_class
 class locked_out(statemachine.State):
     """This device is handled outside of mozpool, and mozpool should not touch it.
@@ -216,57 +266,56 @@ class locked_out(statemachine.State):
 
 
 @DeviceStateMachine.state_class
-class free(AcceptPleaseRequests, statemachine.State):
-    """This device is not in use and available for mozpool requests.  While in this state,
-    Mozpool monitors the device periodically and takes corrective action if it fails."""
+class ready(AcceptPleaseRequests, statemachine.State):
+    """
+    This device is ready for use.  It may be attached to a mozpool request.
+    While in this state, if no request is attached, lifeguard monitors the
+    device periodically and takes corrective action if it fails.
+    """
 
+    # this must be greater than both the ping timeout (10s) and the sut_verify
+    # timeout (195s).  It should be large, too, to keep load on the imaging server
+    # to a reasonable level
     TIMEOUT = 600
 
     def on_entry(self):
-        # TODO: when SUT support is in place, determine whether to use a SUT check or a
-        # ping, or nothing, based on the current image as recorded in the db.
+        # first, if this device is in use by mozpool, don't check its status
+        req_id = self.db.device_requests.get_by_device(self.machine.device_name)
+        if req_id is not None:
+            return
+
+        # otherwise, either sut_verify, or ping, depending on the image capabilities
+        if self.db.devices.has_sut_agent(self.machine.device_name):
+            self.start_sut_verify()
+        else:
+            self.start_ping()
+
+    def start_sut_verify(self):
+        def sut_verified(success):
+            if not success:
+                self.logger.warning("device failed SUT verification")
+                mozpool.lifeguard.driver.handle_event(self.machine.device_name,
+                                                      'failed', {})
+        self.machine.api.sut_verify.start(sut_verified, self.machine.device_name)
+
+    def start_ping(self):
         def ping_complete(success):
             if not success:
-                mozpool.lifeguard.driver.handle_event(self.machine.device_name, 'ping_failed', {})
-        bmm_api.start_ping(self.machine.device_name, ping_complete)
+                self.logger.warning("device failed ping check")
+                mozpool.lifeguard.driver.handle_event(self.machine.device_name,
+                                                      'failed', {})
+        self.machine.api.ping.start(ping_complete, self.machine.device_name)
 
     def on_timeout(self):
-        self.machine.goto_state(free)
+        self.machine.goto_state(ready)
 
-    def on_ping_failed(self, args):
-        self.logger.warning('device stopped pinging while free; running self-test')
+    def on_failed(self, args):
+        # on failure, run a self-test.  Note that self-tests do not end with
+        # 'operation_completed', so they will not accidentally report completion
+        # to mozpool.
         self.machine.goto_state(start_self_test)
 
 
-@DeviceStateMachine.state_class
-class ready(AcceptPleaseRequests, statemachine.State):
-    """
-    This device has been assigned via mozpool and is ready for use by the
-    client.
-    """
-
-    TIMEOUT = 60
-
-    # At one point, this state pinged devices in this state and power-cycled
-    # them when they failed.  This resulted in a lot of unnecessary power cycles
-    # for devices running "flaky" images, with no real benefit.
-
-    def on_free(self, args):
-        self.machine.goto_state(free)
-
-    def on_timeout(self):
-        # It's possible that we can get into this state after a request has
-        # terminated, e.g., if the request is returned but the device is
-        # still booting (since we want to continue through all the states).
-        # Check for that here and return to free if necessary.
-        if data.get_request_for_device(self.machine.device_name):
-            self.machine.goto_state(ready)
-        else:
-            self.logger.warn('in ready state but not assigned to a request; '
-                             'moving to free state')
-            self.machine.goto_state(free)
-
-
 ####
 # Power Cycling
 
@@ -274,10 +323,16 @@ class ready(AcceptPleaseRequests, statemachine.State):
 class pc_power_cycling(PowerCycleMixin, statemachine.State):
     """
     A reboot has been requested, and the device is being power-cycled.  Once
-    the power cycle is successful, go to state 'pc_pinging'.
+    the power cycle is successful, go to state 'sut_verifying'.
     """
 
-    power_cycle_complete_state = 'pc_rebooting'
+    def power_cycle_complete(self):
+        # verify SUT if this device has it; otherwise wait for it to
+        # reboot and then ping it
+        if self.db.devices.has_sut_agent(self.machine.device_name):
+            self.machine.goto_state(sut_verifying)
+        else:
+            self.machine.goto_state(pc_rebooting)
 
 
 @DeviceStateMachine.state_class
@@ -314,7 +369,7 @@ class pc_pinging(statemachine.State):
             # send the machine a power-cycle-ok event on success, and do nothing on failure (timeout)
             if success:
                 mozpool.lifeguard.driver.handle_event(self.machine.device_name, 'ping_ok', {})
-        bmm_api.start_ping(self.machine.device_name, ping_complete)
+        self.machine.api.ping.start(ping_complete, self.machine.device_name)
 
     def on_timeout(self):
         if self.machine.increment_counter('pc_pinging') > self.PERMANENT_FAILURE_COUNT:
@@ -327,89 +382,7 @@ class pc_pinging(statemachine.State):
 
     def on_ping_ok(self, args):
         self.machine.clear_counter('pc_pinging')
-        self.machine.goto_state(sut_verifying)
-
-
-####
-# SUT Verifying
-
-# These states access the device's SUT agent via mozdevice.
-
-@DeviceStateMachine.state_class
-class sut_verifying(statemachine.State):
-    """
-    If the image has a SUT agent, verify that we can establish a connection.
-    If no SUT agent, just proceed to next state.
-    """
-
-    PERMANENT_FAILURE_COUNT = 3
-    TIMEOUT = 45
-
-    def on_entry(self):
-        if not data.device_has_sut_agent(self.machine.device_name):
-            self.on_sut_verify_ok({})
-            return
-        def sut_verified(success):
-            if success:
-                mozpool.lifeguard.driver.handle_event(self.machine.device_name,
-                                                      'sut_verify_ok', {})
-            else:
-                mozpool.lifeguard.driver.handle_event(self.machine.device_name,
-                                                      'sut_verify_failed', {})
-        sut_api.start_sut_verify(self.machine.device_name, sut_verified)
-
-    def on_timeout(self):
-        self.on_sut_verify_failed({})
-
-    def on_sut_verify_failed(self, args):
-        if (self.machine.increment_counter(self.state_name) >
-            self.PERMANENT_FAILURE_COUNT):
-            self.machine.goto_state(failed_sut_verifying)
-        else:
-            self.machine.goto_state(self.state_name)
-
-    def on_sut_verify_ok(self, args):
-        self.machine.clear_counter(self.state_name)
-        self.machine.goto_state(sut_sdcard_verifying)
-
-
-@DeviceStateMachine.state_class
-class sut_sdcard_verifying(statemachine.State):
-    """
-    If the image has a SUT agent, verify that we can write a test file to the
-    device's SD card.
-    If no SUT agent, just proceed to next state.
-    """
-
-    PERMANENT_FAILURE_COUNT = 2
-    TIMEOUT = 210
-
-    def on_entry(self):
-        if not data.device_has_sut_agent(self.machine.device_name):
-            self.on_sut_sdcard_ok({})
-            return
-        def sdcard_verified(success):
-            if success:
-                mozpool.lifeguard.driver.handle_event(self.machine.device_name,
-                                                      'sut_sdcard_ok', {})
-            else:
-                mozpool.lifeguard.driver.handle_event(self.machine.device_name,
-                                                      'sut_sdcard_failed', {})
-        sut_api.start_check_sdcard(self.machine.device_name, sdcard_verified)
-
-    def on_timeout(self):
-        self.on_sut_sdcard_failed({})
-
-    def on_sut_sdcard_failed(self, args):
-        if (self.machine.increment_counter(self.state_name) >
-            self.PERMANENT_FAILURE_COUNT):
-            self.machine.goto_state(failed_sut_sdcard_verifying)
-        else:
-            self.machine.goto_state(self.state_name)
-
-    def on_sut_sdcard_ok(self, args):
-        self.machine.clear_counter(self.state_name)
-        self.machine.goto_state(ready)
+        self.machine.goto_state(operation_complete)
 
 
 ####
@@ -453,12 +426,13 @@ class start_self_test(statemachine.State):
     def on_entry(self):
         self.machine.clear_counter()
         try:
-            data.get_pxe_config_for_device(self.machine.device_name,
+            self.db.devices.get_pxe_config(self.machine.device_name,
                                            'self-test')
-        except data.NotFound:
+        except exceptions.NotFound:
+            # can't self-test this device..
             self.machine.goto_state(pc_power_cycling)
         else:
-            data.set_device_config(self.machine.device_name, 'self-test', '')
+            self.db.devices.set_next_image(self.machine.device_name, 'self-test', '')
             self.machine.goto_state(pxe_power_cycling)
 
 
@@ -471,7 +445,7 @@ class start_maintenance(statemachine.State):
 
     def on_entry(self):
         self.machine.clear_counter()
-        data.set_device_config(self.machine.device_name, 'maintenance', '')
+        self.db.devices.set_next_image(self.machine.device_name, 'maintenance', '')
         self.machine.goto_state(pxe_power_cycling)
 
 
@@ -485,16 +459,15 @@ class pxe_power_cycling(PowerCycleMixin, statemachine.State):
     power_cycle_complete_state = 'pxe_booting'
 
     def setup_pxe(self):
-        # set the pxe config based on what's in the DB
-        cfg = data.device_config(self.machine.device_name)
+        # set the pxe config based on the next image
         try:
-            pxe_config = data.get_pxe_config_for_device(self.machine.device_name)
-        except data.NotFound:
+            pxe_config = self.db.devices.get_pxe_config(self.machine.device_name)
+        except exceptions.NotFound:
             self.logger.warning('no appropriate PXE config found')
             self.machine.goto_state(failed_pxe_booting)
             return
-        bmm_api.set_pxe(self.machine.device_name, pxe_config,
-                        cfg['boot_config'])
+        self.logger.info("setting PXE config to %s" % (pxe_config,))
+        self.machine.api.set_pxe.run(self.machine.device_name, pxe_config)
 
 
 @DeviceStateMachine.state_class
@@ -513,7 +486,7 @@ class pxe_booting(statemachine.State):
 
     def on_mobile_init_started(self, args):
         self.machine.clear_counter(self.state_name)
-        bmm_api.clear_pxe(self.machine.device_name)
+        self.machine.api.clear_pxe.run(self.machine.device_name)
         self.machine.goto_state(mobile_init_started)
 
     def on_timeout(self):
@@ -538,23 +511,28 @@ class mobile_init_started(statemachine.State):
 
     def on_android_downloading(self, args):
         self.machine.clear_counter(self.state_name)
-        bmm_api.clear_pxe(self.machine.device_name)
+        self.machine.api.clear_pxe.run(self.machine.device_name)
         self.machine.goto_state(android_downloading)
 
     def on_b2g_downloading(self, args):
         self.machine.clear_counter(self.state_name)
-        bmm_api.clear_pxe(self.machine.device_name)
+        self.machine.api.clear_pxe.run(self.machine.device_name)
         self.machine.goto_state(b2g_downloading)
 
     def on_maint_mode(self, args):
         # note we do not clear the PXE config here, so it will
-        # continue to boot into maintenance mode
+        # continue to boot into maintenance mode on subsequent reboots
         self.machine.clear_counter(self.state_name)
+        # assume that maintenance has clobbered the image
+        self.db.devices.set_image(self.machine.device_name, None, None)
         self.machine.goto_state(maintenance_mode)
 
     def on_self_test_running(self, args):
         self.machine.clear_counter(self.state_name)
-        bmm_api.clear_pxe(self.machine.device_name)
+        self.machine.api.clear_pxe.run(self.machine.device_name)
+        # assume that the self test has clobbered the image
+        self.db.devices.set_image(self.machine.device_name, None, None)
+        self.db.devices.set_next_image(self.machine.device_name, None, None)
         self.machine.goto_state(self_test_running)
 
     def on_timeout(self):
@@ -575,13 +553,14 @@ class android_downloading(statemachine.State):
     """
 
     # Downloading takes about 30s.  Allow a generous multiple of that to handle
-    # moments of high load or bigger images.  Failures here are likely a misconfig, 
+    # moments of high load or bigger images.  Failures here are likely a misconfig,
     # so they aren't retried very many times
     TIMEOUT = 180
     PERMANENT_FAILURE_COUNT = 3
 
     def on_android_extracting(self, args):
         self.machine.clear_counter(self.state_name)
+        self.db.devices.set_image(self.machine.device_name, None, None)
         self.machine.goto_state(android_extracting)
 
     def on_timeout(self):
@@ -596,7 +575,7 @@ class android_extracting(statemachine.State):
     """
     The second-stage script is extracting the Android artifacts onto the
     sdcard.  When this is complete, the script will send an event and go into
-    the 'android_rebooting' state.
+    the 'sut_verifying' state where it will wait for SUT to come up.
     """
 
     # Extracting takes 2m on a good day.  Give it plenty of leeway.  Funky
@@ -607,7 +586,7 @@ class android_extracting(statemachine.State):
 
     def on_android_rebooting(self, args):
         self.machine.clear_counter(self.state_name)
-        self.machine.goto_state(android_rebooting)
+        self.machine.goto_state(sut_verifying)
 
     def on_timeout(self):
         if self.machine.increment_counter(self.state_name) > self.PERMANENT_FAILURE_COUNT:
@@ -616,62 +595,6 @@ class android_extracting(statemachine.State):
             self.machine.goto_state(pxe_power_cycling)
 
 
-@DeviceStateMachine.state_class
-class android_rebooting(statemachine.State):
-    """
-    The second-stage script has rebooted the device.  This state waits a short
-    time, then begins pinging the device.  The wait is to avoid a false positive
-    ping from the board *before* it has rebooted.
-    """
-
-    # The u-boot loader is also pingable, so we want to be very sure we don't get
-    # a false positive from it.  So, this is quite a bit longer than strictly
-    # necessary, but gives enough time to reboot, run u-boot, and then start Android
-    TIMEOUT = 120
-
-    def on_timeout(self):
-        self.machine.goto_state(android_pinging)
-
-
-@DeviceStateMachine.state_class
-class android_pinging(statemachine.State):
-    """
-    A reboot has been requested, and the power cycle is complete.  Ping until
-    successful, then go to state 'ready'
-    """
-
-    # ping every 10s, failing 12 times (2 minutes total).  If the whole process fails here
-    # a few times, then most likely the image is bogus, so that's a permanent failure.
-    TIMEOUT = 10
-    PING_FAILURES = 12
-    PERMANENT_FAILURE_COUNT = 4
-
-    def on_entry(self):
-        def ping_complete(success):
-            # send the machine a power-cycle-ok event on success, and do nothing on failure (timeout)
-            if success:
-                mozpool.lifeguard.driver.handle_event(self.machine.device_name, 'ping_ok', {})
-        bmm_api.start_ping(self.machine.device_name, ping_complete)
-
-    def on_timeout(self):
-        # this is a little tricky - android_pinging tracks a few tries to ping this device, so
-        # when that counter expires, then we assume the device has not imaged correctly and
-        # retry; *that* failure is counted against PERMANENT_FAILURE_COUNT.
-        if self.machine.increment_counter('ping') > self.PERMANENT_FAILURE_COUNT:
-            self.machine.clear_counter('ping')
-            if self.machine.increment_counter(self.state_name) > self.PERMANENT_FAILURE_COUNT:
-                self.machine.goto_state(failed_android_pinging)
-            else:
-                self.machine.goto_state(pxe_power_cycling)
-        else:
-            self.machine.goto_state(android_pinging)
-
-    def on_ping_ok(self, args):
-        self.machine.clear_counter(self.state_name)
-        self.machine.clear_counter('ping')
-        self.machine.goto_state(sut_verifying)
-
-
 ####
 # PXE Booting :: B2G Installation
 
@@ -683,13 +606,14 @@ class b2g_downloading(statemachine.State):
     """
 
     # Downloading takes about 30s.  Allow a generous multiple of that to handle
-    # moments of high load or bigger images.  Failures here are likely a misconfig, 
+    # moments of high load or bigger images.  Failures here are likely a misconfig,
     # so they aren't retried very many times
     TIMEOUT = 180
     PERMANENT_FAILURE_COUNT = 3
 
     def on_b2g_extracting(self, args):
         self.machine.clear_counter(self.state_name)
+        self.db.devices.set_image(self.machine.device_name, None, None)
         self.machine.goto_state(b2g_extracting)
 
     def on_timeout(self):
@@ -704,7 +628,7 @@ class b2g_extracting(statemachine.State):
     """
     The second-stage script is extracting the B2G artifacts onto the
     sdcard.  When this is complete, the script will send an event and go into
-    the 'b2g_rebooting' state.
+    the 'sut_verifying' state to make sure the image comes up.
     """
 
     # Extracting takes 1m on a good day.  Give it plenty of leeway.  Funky
@@ -715,7 +639,7 @@ class b2g_extracting(statemachine.State):
 
     def on_b2g_rebooting(self, args):
         self.machine.clear_counter(self.state_name)
-        self.machine.goto_state(b2g_rebooting)
+        self.machine.goto_state(sut_verifying)
 
     def on_timeout(self):
         if self.machine.increment_counter(self.state_name) > self.PERMANENT_FAILURE_COUNT:
@@ -724,66 +648,140 @@ class b2g_extracting(statemachine.State):
             self.machine.goto_state(pxe_power_cycling)
 
 
-@DeviceStateMachine.state_class
-class b2g_rebooting(statemachine.State):
-    """
-    The second-stage script has rebooted the device.  This state waits a short
-    time, then begins pinging the device.  The wait is to avoid a false positive
-    ping from the board *before* it has rebooted.
-    """
-
-    # The u-boot loader is also pingable, so we want to be very sure we don't get
-    # a false positive from it.  So, this is quite a bit longer than strictly
-    # necessary, but gives enough time to reboot, run u-boot, and then start b2g.
-    TIMEOUT = 120
-
-    def on_timeout(self):
-        self.machine.goto_state(b2g_pinging)
+####
+# SUT Verifying
 
+# This sequence of states checks the device using SUT.  It forms the final set
+# of operations for several sequences above, and goes to the 'ready' state when
+# it is finished.
 
 @DeviceStateMachine.state_class
-class b2g_pinging(statemachine.State):
+class sut_verifying(statemachine.State):
     """
-    A reboot has been requested, and the power cycle is complete.  Ping until
-    successful, then go to state 'ready'
+    Verify that we can establish a connection to the device's SUT agent.
+    This assumes that the device has a SUT agent on it.
     """
 
-    # TODO: factor out into a mixin to combine with android_pinging
-
-    # ping every 10s, failing 12 times (2 minutes total).  If the whole process fails here
-    # a few times, then most likely the image is bogus, so that's a permanent failure.
-    TIMEOUT = 10
-    PING_FAILURES = 12
-    PERMANENT_FAILURE_COUNT = 4
+    # wait a bit over 20m total for the device to reboot and come up
+    PERMANENT_FAILURE_COUNT = 30
+    # power-cycling every 450s = 7.3m
+    POWER_CYCLE_EVERY = 10
+    # the SUT check can take up to 45s, so don't reduce this further
+    TIMEOUT = 45
 
     def on_entry(self):
-        def ping_complete(success):
-            # send the machine a power-cycle-ok event on success, and do nothing on failure (timeout)
+        def sut_verified(success):
             if success:
-                mozpool.lifeguard.driver.handle_event(self.machine.device_name, 'ping_ok', {})
-        bmm_api.start_ping(self.machine.device_name, ping_complete)
+                mozpool.lifeguard.driver.handle_event(self.machine.device_name,
+                                                      'sut_verify_ok', {})
+             # if not, wait for the timeout to occur, rather than immediately
+             # re-checking
+        self.machine.api.sut_verify.start(sut_verified, self.machine.device_name)
 
     def on_timeout(self):
-        # this is a little tricky - b2g_pinging tracks a few tries to ping this device, so
-        # when that counter expires, then we assume the device has not imaged correctly and
-        # retry; *that* failure is counted against PERMANENT_FAILURE_COUNT.
-        if self.machine.increment_counter('ping') > self.PERMANENT_FAILURE_COUNT:
-            self.machine.clear_counter('ping')
-            if self.machine.increment_counter(self.state_name) > self.PERMANENT_FAILURE_COUNT:
-                self.machine.goto_state(failed_b2g_pinging)
-            else:
-                self.machine.goto_state(pxe_power_cycling)
+        ctr = self.machine.increment_counter(self.state_name)
+        if ctr > self.PERMANENT_FAILURE_COUNT:
+            self.machine.goto_state(failed_sut_verifying)
+        elif ctr % self.POWER_CYCLE_EVERY == 0:
+            self.machine.goto_state(sut_verify_power_cycle)
         else:
-            self.machine.goto_state(b2g_pinging)
+            self.machine.goto_state(self.state_name)
 
-    def on_ping_ok(self, args):
+    def on_sut_verify_ok(self, args):
         self.machine.clear_counter(self.state_name)
-        self.machine.clear_counter('ping')
-        self.machine.goto_state(sut_verifying)
+        self.machine.goto_state(sut_sdcard_verifying)
+
+
+@DeviceStateMachine.state_class
+class sut_verify_power_cycle(statemachine.State):
+    """
+    sut_verifying has failed a few times.  Power-cycle the device and try again.
+    """
+
+    PERMANENT_FAILURE_COUNT = 3
+    TIMEOUT = 60                    # must be greater than 42s; see above
+
+    def on_entry(self):
+        def powercycle_done(success):
+            # send the machine a power-cycle-ok event on success, and do nothing on failure (timeout)
+            if success:
+                mozpool.lifeguard.driver.handle_event(self.machine.device_name, 'power_cycle_ok', {})
+        self.machine.api.powercycle.start(powercycle_done, self.machine.device_name)
+
+    def on_power_cycle_ok(self, args):
+        self.machine.goto_state('sut_verifying')
+
+    def on_timeout(self):
+        if self.machine.increment_counter(self.state_name) > self.PERMANENT_FAILURE_COUNT:
+            self.machine.goto_state(failed_sut_verifying)
+        else:
+            self.machine.goto_state(self.state_name)
+
+@DeviceStateMachine.state_class
+class sut_sdcard_verifying(statemachine.State):
+    """
+    Verify that we can write a test file to the device's SD card.
+    """
+
+    PERMANENT_FAILURE_COUNT = 2
+    # this should be greater than the sdcard timeout (30s)
+    TIMEOUT = 60
+
+    def on_entry(self):
+        def sdcard_verified(success):
+            if success:
+                mozpool.lifeguard.driver.handle_event(self.machine.device_name,
+                                                      'sut_sdcard_ok', {})
+            else:
+                mozpool.lifeguard.driver.handle_event(self.machine.device_name,
+                                                      'sut_sdcard_failed', {})
+        self.machine.api.check_sdcard.start(sdcard_verified, self.machine.device_name)
+
+    def on_timeout(self):
+        self.on_sut_sdcard_failed({})
+
+    def on_sut_sdcard_failed(self, args):
+        if (self.machine.increment_counter(self.state_name) >
+            self.PERMANENT_FAILURE_COUNT):
+            self.machine.goto_state(failed_sut_sdcard_verifying)
+        else:
+            self.machine.goto_state(self.state_name)
+
+    def on_sut_sdcard_ok(self, args):
+        self.machine.clear_counter(self.state_name)
+        self.machine.goto_state(operation_complete)
+
+
+@DeviceStateMachine.state_class
+class operation_complete(ImagingResultMixin, statemachine.State):
+    """
+    An intermediate state to wrap up an operation.  This takes care of some
+    housekeeping before returning to 'ready'
+    """
+
+    TIMEOUT = 60
+
+    def on_entry(self):
+        # if there's a next_image, then the imaging is now complete, so set the
+        # image and blank out the next image
+        device_name = self.machine.device_name
+        new_img = self.db.devices.get_next_image(device_name)
+        if new_img['image']:
+            self.db.devices.set_image(device_name,
+                    new_img['image'], new_img['boot_config'])
+            self.db.devices.set_next_image(device_name, None, None)
+
+        # inform mozpool, if it cares
+        self.send_imaging_result('complete')
+        self.machine.goto_state(ready)
+
+    def on_timeout(self):
+        # re-enter this state in case we get stuck here by a server failure
+        self.machine.goto_state(operation_complete)
 
 
 ####
-# PXE Booting :: Self-Test
+# PXE Booting :: Maintenance mode
 
 @DeviceStateMachine.state_class
 class maintenance_mode(AcceptBasicPleaseRequests, statemachine.State):
@@ -795,7 +793,7 @@ class maintenance_mode(AcceptBasicPleaseRequests, statemachine.State):
 
 
 ####
-# PXE Booting :: Maintenance Mode
+# PXE Booting :: Self-Test
 
 @DeviceStateMachine.state_class
 class self_test_running(AcceptBasicPleaseRequests, statemachine.State):
@@ -813,23 +811,29 @@ class self_test_running(AcceptBasicPleaseRequests, statemachine.State):
         self.machine.goto_state(failed_self_test)
 
     def on_self_test_ok(self, args):
-        self.machine.goto_state(free)
+        self.machine.goto_state(ready)
 
 ####
 # Failure states
 
-class failed(AcceptBasicPleaseRequests, statemachine.State):
+class failed(AcceptBasicPleaseRequests, ImagingResultMixin, statemachine.State):
     """Parent class for failed_.. classes.  The only way out is a self-test or
     maintenance mode."""
 
+    # by default, a failure will be indicated to mozpool with this imaging result.
+    # subclasses can set this for other results
+    imaging_result = 'failed-bad-device'
+
+    # if true, this state should bounce immediately into a self-test operation
+    try_self_test = False
+
     def on_entry(self):
         self.logger.error("device has failed: %s" % self.state_name)
+        self.send_imaging_result(self.imaging_result)
+        if self.try_self_test:
+            self.machine.goto_state(start_self_test)
 
 
-@DeviceStateMachine.state_class
-class failed_imaging(failed):
-    "The imaging process could not be started"
-
 @DeviceStateMachine.state_class
 class failed_power_cycling(failed):
     "The power-cycle operation itself has failed or timed out multiple times"
@@ -845,34 +849,42 @@ class failed_mobile_init_started(failed):
 @DeviceStateMachine.state_class
 class failed_sut_verifying(failed):
     "Could not connect to SUT agent."
+    try_self_test = True
 
 @DeviceStateMachine.state_class
 class failed_sut_sdcard_verifying(failed):
     "Failed to verify device's SD card."
+    try_self_test = True
 
 @DeviceStateMachine.state_class
 class failed_android_downloading(failed):
     "While installing Android, the device timed out repeatedly while downloading Android"
+    try_self_test = True
 
 @DeviceStateMachine.state_class
 class failed_android_extracting(failed):
     "While installing Android, the device timed out repeatedly while extracting Android"
+    try_self_test = True
 
 @DeviceStateMachine.state_class
 class failed_android_pinging(failed):
     "While installing Android, the device timed out repeatedly while pinging the new image waiting for it to come up"
+    try_self_test = True
 
 @DeviceStateMachine.state_class
 class failed_b2g_downloading(failed):
     "While installing B2G, the device timed out repeatedly while downloading B2G"
+    try_self_test = True
 
 @DeviceStateMachine.state_class
 class failed_b2g_extracting(failed):
     "While installing B2G, the device timed out repeatedly while extracting B2G"
+    try_self_test = True
 
 @DeviceStateMachine.state_class
 class failed_b2g_pinging(failed):
     "While installing B2G, the device timed out repeatedly while pinging the new image waiting for it to come up"
+    try_self_test = True
 
 @DeviceStateMachine.state_class
 class failed_self_test(failed):
